;;;===================================================
;;; OM-CHANT
;;; Control CHANT synthesis from OpenMusic
;;;
;;; CHANT Synthesis / Normalization
;;; Jean Bresson, IRCAM 2010
;;;===================================================


(in-package :om)

(defvar *chant-path* nil)
(defvar *snd-path* nil)
(defvar *tosf-path* nil)

(defun snd-normalize (infile &optional outfile gain resolution)
  (unless *snd-path* 
    (setf *snd-path* (make-pathname :directory (pathname-directory *chant-path*)
                                    :name "snd")))
  (if (probe-file *snd-path*)
      (let ((out (cond ((pathnamep outfile) outfile)
                       ((stringp outfile) (outfile outfile))
                       (t (om-choose-new-file-dialog :directory (outfile nil)))))
            (res (if (and resolution (numberp resolution)) resolution  *audio-res*))
            (val (or gain (if (numberp *normalize-level*) (db->lin *normalize-level*) 1.0))))
        
        (print "==========================")
        (print "SND: NORMALISATION")
        (print "==========================")
        
        (om-cmd-line 
         (string+ (format nil "PATH=$PATH:~s; " (namestring (make-pathname :directory (pathname-directory *CHANT-PATH*))))
                  (format nil "~s -aNormalize -i~s -o~s -fAiff -d~D -n2" 
                          (om-path2cmdpath *snd-path*)
                          (om-path2cmdpath infile) 
                          (om-path2cmdpath outfile)
                          res)) *sys-console*)
         (or (probe-file (om-namestring outfile))
             (om-beep-msg "Error in SND normalization!!")))
        (om-beep-msg "SND tool not found!!")))


(defun tosf-convert (infile outfile nch res sr)
  (if (not (probe-file *tosf-path*))
      (progn
        (print "TOSF not found!!!")
        (print "TOSF is a utility for converting the CHANT output data to a sound file.")
        (print "Please check TOSF path in the Externals preferences")
        nil)
    (progn
      (print "==========================")
      (print "TOSF: OUTPUT TO AIFF")
      (print "==========================")
  
      (om-cmd-line (print (format nil "~s -c~D -of -R~D ~s < ~s" 
                                  (namestring *tosf-path*)
                                  nch sr
                                  (namestring outfile)
                                  (namestring infile)
                                  ))
                   *sys-console*)
  
      (probe-file outfile))))
  

(defmethod! chant-synth ((self t) &key outfile resolution normalize-level)
    (om-beep-msg "INVALID INPUT TO CHANT-SYNTH!!!"))

(defmethod! chant-synth ((self pathname) &key outfile resolution normalize-level)
  :icon '(410)
  :indoc '("SDIFfile object or file pathname" "output file pathname" "output sample resolution")
  :initvals '(nil nil nil)
  :doc "Calls CHANT to synthesize a sound from <self>.

<self> is and SDIF file formatted for the control of CHANT (see MAKE-CHANT-SDIF-FILE).

Note : the CHANT path must be specified in the OM preferences.

<outfile> is an output file pathname. If NIL, the name  of <self> will be used with a sound file extension and the output will be written in the OM output files folder.

CHANT output raw data files, which are normalized internally by CHANT-SYNTH.
Some normalization modules may be required (check messages in the OM Listener!)

<resolution> sets the resolution for the raw data -> sound file conversion.
"
  (let* ((path-sdif self)
         (outpath (handle-new-file-exists 
                   (corrige-sound-filename (if outfile outfile (string+ (pathname-name self) ".aiff")) *om-outfiles-folder*)))
         (tmppath1 (handle-new-file-exists
                    (make-pathname :device (pathname-device outpath) :directory (pathname-directory outpath)
                                   :name (pathname-name outpath) :type "tmp1")))
         (tmppath2 (handle-new-file-exists
                    (make-pathname :device (pathname-device outpath) :directory (pathname-directory outpath)
                                  :name (pathname-name outpath) :type "sf")))
         (res (or resolution *audio-res*)))
  (if (not (probe-file *CHANT-PATH*))
      (om-beep-msg "CHANT not found !!")
      (if (not (probe-file path-sdif))
          (om-beep-msg "ERROR : bad INPUT file pathname !!")
        (let ((nch (find-in-nvtlist (getnvtlist path-sdif) "NumberOfChannels"))
              (sr (find-in-nvtlist (getnvtlist path-sdif) "SampleRate")))
          (print "==========================")
          (print "CHANT SYNTHESIS")
          (print "==========================")
          (print (format nil "Running CHANT with ~D" (namestring path-sdif)))
          (setf nch (or (and nch (read-from-string nch)) 1)
                sr (or (and sr (read-from-string sr)) *audio-sr*))
          (print (format nil "Number of channels= ~D" nch))
          (om-cmd-line 
           (string+ ;(format nil "PATH=$PATH:~s; " (namestring (make-pathname :directory (pathname-directory *CHANT-PATH*))))
                    (format nil "cd ~s; " (namestring (make-pathname :directory (pathname-directory *CHANT-PATH*))))
                    (om-print (format nil "~s -v -i~s -o~s" 
                                      (om-path2cmdpath *CHANT-PATH*)
                                      (om-path2cmdpath path-sdif) 
                                      (om-path2cmdpath tmppath1)))
                    ) 
           *sys-console*)
        
          
        
          (if (not (probe-file tmppath1))
              (om-beep-msg "!! ERROR in CHANT synthesis !!")
            (progn
             (tosf-convert tmppath1 tmppath2 nch res sr)
             ;(snd-normalize tmppath1 tmppath2 0.0 res)
             ;(csound-normalize tmppath1 tmppath2 -3.0 res)
             ;(setf tmppath2 tmppath1)
            
             (add-tmp-file tmppath1)

             (if (not (probe-file tmppath2))
                 (om-beep-msg "ERROR in TOSF conversion !!")
               ; (snd-normalize tmppath2 outpath 0.0 res)
               ;(tosf-convert tmppath2 outpath nch res sr)
               (if (require-library "OM-SuperVP" nil)
                   (progn 
                     (unless (and *svp-path* (probe-file *svp-path*))
                       (om-message-dialog (format nil "SuperVP is required to normalize the Chant output.~%~%The library OM-SuperVP was successfully loaded, but the supervp executable could not be found. Please indicate the supervp path in the next dialog.~%~%In order to record the supervp path, please go to the OM Preferences / Externals."))
                       (let ((exepath (om-choose-file-dialog :prompt "Set path to SuperVP...")))
                         (when (and exepath (probe-file exepath))
                           (setq *svp-path* exepath))))
                     (supervp-normalize tmppath2 outpath (or normalize-level *normalize-level* 0) res)
                     (add-tmp-file tmppath2))
                 (progn
                   (om-beep-msg "!! Warning: OM-SuperVP not found for CHANT output normalization... !!")
                   (setf outpath tmppath2)))
               )
             
             (when *delete-inter-file* (clean-tmp-files))
             (and outpath (probe-file outpath))
             ))
          ))
      )))

(defmethod! chant-synth ((self sdiffile) &key outfile resolution normalize-level)
   (chant-synth (filepathname self) :outfile outfile :resolution resolution :normalize-level normalize-level))

(defmethod! chant-synth ((self string) &key outfile resolution normalize-level)
   (if (probe-file (pathname self))
       (chant-synth (pathname self) :outfile outfile :resolution resolution :normalize-level normalize-level)
     (om-beep-msg (string+ "File " self " not found!"))))




;=================================================


(defmethod! transitions ((tmpobjlist list) mkev mkinter modifbeg modifend)
  (let ((sortedlist (sort tmpobjlist '< :key 'offset)))
    (flat (loop for curr-ev in sortedlist
                for prev-ev in (x-append (car sortedlist) sortedlist)
                for next-ev in (x-append (cdr sortedlist) (last-elem sortedlist))
                collect 
                (if 
                          (and (< (offset prev-ev) (offset curr-ev)) ; secu begin
                               (> (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev))) (offset curr-ev)))
                          ;;; p2 > e1
                          (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                   (> (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                            ;;; e2 > n1
                            ;;;'double-switch
                            (let ((evt (funcall mkev curr-ev)))
                              (funcall modifend 
                                       (funcall modifbeg evt (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev))))
                                       (offset next-ev))
                              )
                            (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                     (< (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                              ;;; e2 < n1
                              ;;;'begin-switch+silence
                              (let ((evt (funcall mkev curr-ev))
                                    (nex (funcall mkev next-ev)))
                                (list 
                                 (funcall modifbeg evt (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev)))) 
                                 (funcall mkinter evt nex)))
                              ;;;'begin-switch
                              (let ((evt (funcall mkev curr-ev)))
                                (funcall modifbeg evt (+ (offset prev-ev) (* (extend prev-ev) (strech-fact prev-ev))))
                                )
                              )
                            )
                          (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                   (> (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                            ;;; e2 > n1
                            ;;;'end-switch
                            (let ((evt (funcall mkev curr-ev)))
                              (funcall modifend evt (offset next-ev))
                              )
                            (if (and (< (offset curr-ev) (offset next-ev)) ; secu end
                                     (< (+ (offset curr-ev) (* (extend curr-ev) (strech-fact curr-ev))) (offset next-ev)))
                              ;;; e2 < n1
                              ;;; 'evt+silence
                              (let ((evt (funcall mkev curr-ev))
                                    (nex (funcall mkev next-ev)))
                                (list evt (funcall mkinter evt nex))
                                )
                              ;;; 'evt
                              (let ((evt (funcall mkev curr-ev)))
                                evt)
                              )
                            )
                          )
                ))))
  









