;;;===========================================================================
;;; OM-Spat
;;; Rendering of spatial scene descriptions 
;;; J. Bresson (IRCAM - 2010)
;;;===========================================================================

;;;===========================================================================
;;; spat renderer interface
;;;
;;; Requires perliminary installation of spatdif.renderer (protected Ircam)
;;; Spat renderer by T. Carpentier (IRCAM - 2010)
;;;=============================================================================

(in-package :om)

;;;===========================

(defmethod get-speakers-file ((self pathname)) self)
(defmethod get-speakers-file ((self string)) (pathname self))

(defmethod get-speakers-file ((self list)) 
  (let ((spkfile (tmpfile "speakers.txt")))
    (save-params (cons '("xyz") self) spkfile)))

(defmethod get-speakers-file ((self textfile)) 
  (get-speakers-file (list-of-data (buffer-text self))))

(defmethod get-speakers-file ((self 3DC)) 
  (get-speakers-file (point-pairs self)))

(defmethod get-speakers-file ((self BPC)) 
  (get-speakers-file (mapcar #'(lambda (2Dp) (append 2Dp (list 0))) (point-pairs self))))


;;;===========================

(defmethod! spat-synth ((self string) out-config &key out-path speakers-pos hrtf)
   :icon '(410)
   :initvals '(nil nil nil nil nil)
   :indoc '("sdif file or spat-matrix" "number of speakers" "output file pathname" "speakers position" "HRTF file")
   :outdoc '("sound file pathname")
   :doc "Synthesizes a multichannel sound file from a SPAT-MATRIX or an SDIF file using Spat 4.

Requires preliminary osntallation of Spat and path setup in OM prefernces.

<self> can be a pathname to the SDIF file, and SDIFFILE object, or a SPAT-MATRIX.
<out-config> can be a integer (number of speakers) or 'binaural for binaral rendering.
<out-path> is the pathname of the file to create. If not specified, will be asked by a file chooser dialog.
<speaker-pos> is the speakers configuration (list of positions in 3D - xyz), provided as a BPC, 3DC, as a list, a textfile or a text file pathname. 
If not provided, default positions are used.
<hrtf> is tha path to an HRTF file for binaural rendering. A default HRTF is used if not specified.

Spat by Thibaut Carpentier, Ircam Acoustic/Cognitive Spaces, 2010.
"

   (let* ((outfile (handle-new-file-exists (or out-path (om-choose-new-file-dialog :types (list (om-str :all-files) "*.*" 
                                                                                                (format nil (om-str :file-format) "AIFF") "*.aiff" 
                                                                                                (format nil (om-str :file-format) "WAV") "*.wav")))))

          (mode (cond ((equal out-config 'binaural) "binaural")
                      (t *spat-default-panning*)))
          (nch (cond ((equal out-config 'binaural) 2)
                     ((numberp out-config) out-config)
                     (t *spat-default-numout*)))
          (nin (length (sdifstreams (load-sdif-file self))))
          (hrtffile (if (string-equal mode "binaural") (string+ " -H \"" (namestring (or (and hrtf (probe-file hrtf))
                                                                                         *spat-default-hrtf*)) "\"") 
                      ""))
          (speakers "")
          (decoding (if (stringp *spat-default-decoding*) (string+ " -d " *spat-default-decoding*) ""))
          (buffersize (or *spat-default-buffersize* 512))
          (nrev (or *spat-default-nreverb* 1))
          (outformat (if (equal *def-snd-format* 'wav) "wav" "aiff"))
          (nspatchannels (or *spat-default-nchannels* 8)))
     (when outfile
       (setf outfile (om-make-pathname :directory outfile :name (pathname-name outfile)
                                       :type outformat)) ;;; because Spat always create "aiff" extension (not "aif")
       (when speakers-pos 
         (let ((speakers-file (get-speakers-file speakers-pos)))
           (when (probe-file speakers-file) 
             (add-tmp-file speakers-file)
             (setf speakers (string+ " -S " (namestring speakers-file) "")))))
     
       (if (probe-file *spat-renderer*)
           (om-cmd-line (format nil "~s -p ~A ~A -i ~D -o ~D -r ~D -I ~D -M 1 -f ~s -D ~s -s ~s -O ~s -F ~s -v 1 -b ~D ~A ~A"
                                (namestring *spat-renderer*)
                                mode decoding nin nch nrev nspatchannels
                                self
                                (namestring (om-make-pathname :directory outfile))
                                (namestring (om-make-pathname :directory outfile))
                                (pathname-name outfile)
                                outformat
                                buffersize
                                hrtffile
                                speakers)
                        *sys-console*)
         (om-beep-msg "Spat renderer not found. Set path in OM Preferences !!!"))
       (when *delete-inter-file* (clean-tmp-files)) 
       (and outfile (probe-file outfile)))))

(defmethod! spat-synth ((self pathname) out-config &key out-path speakers-pos hrtf)
    (spat-synth (namestring self) out-config :out-path out-path :speakers-pos speakers-pos :hrtf hrtf))

(defmethod! spat-synth ((self sdiffile) out-config &key out-path speakers-pos hrtf)
    (spat-synth (filepathname self) out-config :out-path out-path :speakers-pos speakers-pos :hrtf hrtf))

(defmethod! spat-synth ((self spat-matrix) out-config &key out-path speakers-pos hrtf)
   (let* ((tmppath (handle-new-file-exists (tmpfile "spat.sdif")))
          (sdif (save-spat-sdif self :stream-mode 'sep :out tmppath :export-sounds :temp)))
     (add-tmp-file tmppath)
     (spat-synth sdif out-config :out-path out-path :speakers-pos speakers-pos :hrtf hrtf)
     ))

#|
spatdif.renderer options:          
spatdif.renderer options :         
   -f <sdif file fullpath>         
   -D <destination folder>         
   -O <output name>                            (default : spat_out)
   -F <output format>                          (default : aiff)
   -i <number of sources/inputs>               (default : 1)
   -o <number of speakers/outputs>             (default : 8)
   -I <number of internal channels>            (default : 8)
   -r <number of reverbs>                      (default : 1)
   -b <buffersize>                             (default : 512)
   -v <verbose>                                (default : 0)
   -M <multichannel output file>               (default : 0)
   -p <panning type>                           (default : angular)
       [ binaural, xy, ms, ab, surround, panr, angular, vbap2d, vbap, dbap, bformat, hoa2d, hoa3d, null ]
   -d <decoding type>                          (default : null)
       [ null, transaural, bformat, hoa2d, hoa3d, surround, bformat2uhj, uhj2bformat ]
   -H <hrtf file fullpath>      
   -S <speakers file>              
   File format is as follow:
    xyz
    1 -1 1 
    0.7 -0.7 0
    1 2 3

Ex: ./spatdif.renderer -p angular -i 1 -o 4 -d null -f ../test/spat-test-simple.sdif -r 1 -D ../test -v 1 -b 32 -M 1 -S ../test/speakers.txt -F aiff -O mySpatOutput
|#
