;;;===================================================
;;;
;;; OM-SuperVP
;;; SuperVP sound analysis and processing for OpenMusic
;;;
;;; Requires perliminary installation of SuperVP (also included in AudioSculpt)
;;; Set SuperVP path in OM preferences (externals) once the library is loaded.
;;;
;;;
;;; SOUND ANALYSIS TOOLS
;;; Authors: Jean Bresson (IRCAM - 2006-2010)
;;;
;;;===================================================


(in-package :om)


;;; AUTO UPDATE COMPATIBILITY

(defun cdr-inputs-to-keywords (fun inputs)  
  (cons (car inputs) 
        (loop for input in (cdr inputs)
              for argname in (cdr (ordered-arg-list fun))
              collect (if (keyword-input-p input) input
                        (make-instance 'input-keyword
                                       :name (name input) 
                                       :value argname
                                       :def-value (value input)
                                       :doc-string (doc-string input)
                                       :val-menu (if (menu-input-p input) (thepopup input) nil)
                                       :box-ref (box-ref input)
                                       :connected? (connected? input)
                                       )))))


(defmethod update-inputs ((self (eql 'fft)) inputs)
  (cdr-inputs-to-keywords self inputs))

(defmethod update-inputs ((self (eql 'transient-detection)) inputs)
  (cdr-inputs-to-keywords self inputs))

(defmethod update-inputs ((self (eql 'f0-estimate)) inputs)
  (cdr-inputs-to-keywords self inputs))




;;;================================================================================================================
;;; FFT
;;;================================================================================================================

(defmethod! fft ((infile string) &key begin-time end-time (windowsize 4096) (fftsize 4096) (step 256) 
                 (windowtype "hanning") (gabarit 1) (out "fft.sdif"))
  :initvals '(nil nil nil 4096 4096 256 "hanning" 1 "fft.sdif")
  :menuins (list (list 6 '(("Blackman" "blackman") ("Hanning" "hanning") ("Hamming" "hamming")))
                 (list 7 '(("log amplitude" 0) ("amplitude" 1) ("phase" 2) ("phase / amplitude" 3) 
                           ("frequency / amplitude" 4) ("real / imaginary" 5))))
  :icon 952
  :doc "Calculates the FFT of a Sound file using SuperVP.
The results of analysis are stored in an SDIF file which pathname is returned.

- <infile> : pathname or SOUND object to be analysed

- <begin-time> : begin time of the analysis

- <end-time> : end time of the analysis

- <windowsize> : number of samples of the analysis window

- <fftsize> : number of points in fft

- <step> : number of samples between two successive analysis windows

- <windowtype> : shape of the analysis window

- <gabarit> : output fft format
            0 = logarithmic amplitudes
            1 = amplitude
            2 = phase
            3 = phase, amplitude
            4 = frequency, amplitude
            5 = real, imaginary 
Note that only sdiffiles at gabarit 3 and 5 can be used to recalculate a sound file using IFFT.

- <out> : output file pathname
"

  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let ((outname (if out
                      (handle-new-file-exists (if (pathnamep out) out (outfile out)))
                      (om-choose-new-file-dialog :prompt "Choose a new SDIF FFT file"
                                                 :directory (def-save-directory)))))
      (when outname
        (setf *last-saved-dir* (make-pathname :directory (pathname-directory outname)))
        (let* ((unix-outname (om-path2cmdpath outname))
               (beginstr (if begin-time (format nil "-B~D " begin-time) ""))
               (endstr (if end-time (format nil "-E~D " end-time) ""))
               (fftstr (format nil "-N~D -M~D -W~D -I~D " fftsize windowsize windowtype step))
               (ostr (format nil "-OS~D " (if (and gabarit (>= gabarit 0) (<= gabarit 5)) gabarit 
                                            (progn (om-message-dialog (format nil "FFT gabarit ~D does not exist.~% FFT will be formated at gabarit 1" gabarit))
                                              1)))) 
               (cmd (format nil "~s -v -t -S~s -Afft -U ~A ~A ~A ~s ~s" 
                            (om-path2cmdpath *SVP-PATH*)
                            (om-path2cmdpath infile)
                            fftstr beginstr endstr ostr unix-outname))
               )
          (om-print "==========================")
          (om-print "SUPERVP FFT ANALYSIS")
          (om-print "==========================")
          (om-print cmd)        
          (om-cmd-line cmd *sys-console*)
      (and outname (probe-file outname)))))
    (om-beep-msg "SuperVP not found!"))
  )

(defmethod! fft ((infile pathName) &key begin-time end-time (windowsize 4096) (fftsize 4096) (step 256) 
                 (windowtype "hanning") (gabarit 1)  (out "fft.sdif"))
  (fft (om-namestring infile) 
       :begin-time begin-time :end-time end-time 
       :fftsize fftsize :windowsize windowsize
       :step step :windowtype windowtype :gabarit gabarit 
       :out out))

(defmethod! fft ((infile sound) &key begin-time end-time (windowsize 4096) (fftsize 4096) (step 256) 
                 (windowtype "hanning") (gabarit 1) (out "fft.sdif"))
  (fft (sound-path infile) 
       :begin-time begin-time :end-time end-time 
       :fftsize fftsize :windowsize windowsize
       :step step :windowtype windowtype :gabarit gabarit 
       :out out))



;;;================================================================================================================
;;; FUNDAMENTAL FREQUENCY ESTIMATE
;;;================================================================================================================

(defmethod! f0-estimate ((infile string) &key begin-time end-time 
                         (fund-minfreq 50.0) (fund-maxfreq 1000.0) (spectrum-maxfreq 4000.0) (noise-threshold 50.0) (smooth-order 3)
                         (windowsize 4096) (fftsize 4096) (step 256) (windowtype "hanning") (out "f0.sdif"))
  :initvals '(nil nil nil 50.0 1000.0 4000.0 50.0 3 4096 4096 256 "hanning" "f0.sdif")
  :menuins (list (list 11 '(("Blackman" "blackman")("Hanning" "hanning")("Hamming" "hamming"))))
  :icon 952
  :doc "Calculates the Fundamental Fequency in <FileName> using SuperVP.
The results of analysis are stored in an SDIF file which pathname is returned.

- <infile> : pathname or SOUND object to be analysed

- <begin-time> : the begin time of the analysis (s)
 
- <end-time> : the end time of the analysis (s)

- <fund-minfreq> : Fundamental minimal frequency (Hz)

- <fund-maxfreq> : Fundamental maximal frequency (Hz)

- <spectrum-maxfreq> : Maximal frequency in spectrum (Hz)

- <noise-threshold> : Noise threshold (dB)

- <smooth-order> : Smooth Order (int)

- <windowsize> : number of samples of the analysis window

- <fftsize> : number of points of fft

- <step> : number of samples between two successive analysis windows

- <windowtype> : shape of the analysis window

- <out> : output file pathname

"
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let ((outname (if out 
                      (handle-new-file-exists (if (pathnamep out) out (outfile out)))
                      (om-choose-new-file-dialog :prompt "Choose a new SDIF F0 file"
                                                 :directory (def-save-directory)))))
      (when outname
        (setf *last-saved-dir* (make-pathname :directory (pathname-directory outname)))
        (let* ((unix-outname (om-path2cmdpath outname))
              (beginstr (if begin-time (format nil "-B~D " begin-time) ""))
              (endstr (if end-time (format nil "-E~D " end-time) ""))
              (fftstr (format nil "-N~D -M~D -W~D -I~D " fftsize windowsize windowtype step))
              (f0params (format nil "fm~D fM~D F~D sn~D smooth~D " 
                                fund-minfreq fund-maxfreq spectrum-maxfreq noise-threshold smooth-order))
              (cmd (format nil "~s -v -t -ns -U -S~s -Af0 ~s ~A ~A ~A -OS0 ~s" 
                           (om-path2cmdpath *SVP-PATH*)
                           (om-path2cmdpath infile)
                           f0params
                           fftstr beginstr endstr unix-outname))
              
              )
          (om-print "===========================")
          (om-print "SUPERVP F0 ANALYSIS")
          (om-print "===========================")
          (om-print cmd)   
          (om-cmd-line cmd *sys-console*)
          (and outname (probe-file outname)))))
        (om-beep-msg "!!! SuperVP not found !!!"))
      )


(defmethod! f0-estimate ((infile pathName) &key begin-time end-time 
                         (fund-minfreq 50.0) (fund-maxfreq 1000.0) (spectrum-maxfreq 4000.0) (noise-threshold 50.0) (smooth-order 3)
                         (windowsize 4096) (fftsize 4096) (step 256) (windowtype "hanning") (out "f0.sdif"))
  (f0-estimate (om-namestring infile) 
               :begin-time begin-time :end-time end-time 
               :fund-minfreq fund-minfreq :fund-maxfreq fund-maxfreq :spectrum-maxfreq spectrum-maxfreq 
               :noise-threshold noise-threshold :smooth-order smooth-order
               :windowsize windowsize :fftsize fftsize :step step :windowtype windowtype 
               :out out))

(defmethod! f0-estimate ((infile sound) &key begin-time end-time 
                         (fund-minfreq 50.0) (fund-maxfreq 1000.0) (spectrum-maxfreq 4000.0) (noise-threshold 50.0) (smooth-order 3)
                         (windowsize 4096) (fftsize 4096) (step 256) (windowtype "hanning") (out "f0.sdif"))  
  (f0-estimate (sound-path infile) 
               :begin-time begin-time :end-time end-time 
               :fund-minfreq fund-minfreq :fund-maxfreq fund-maxfreq :spectrum-maxfreq spectrum-maxfreq 
               :noise-threshold noise-threshold :smooth-order smooth-order
               :windowsize windowsize :fftsize fftsize :step step :windowtype windowtype 
               :out out))

;;;================================================================================================================
;;; TRANSIENT DETECTION : GENERATE MARKERS
;;;================================================================================================================

(defmethod! transient-detection ((infile string) &key
                                 (treshold 1.4) (minfreq 0.0) (maxfreq 22050.0) (minoffset 0.02)
                                 (windowsize 4096) (fftsize 4096) (step-oversamp 8) (windowtype "hanning") 
                                 (out "markers.sdif"))
  :initvals '(nil 1.4 0.0 22050.0 0.02 4096 4096 8 "hanning" "markers.sdif")
  :menuins (list (list 7 '(("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
                 (list 8 '(("Blackman" "blackman")("Hanning" "hanning")("Hamming" "hamming"))))
  :icon 952
  :doc "Detects transients in <filename> using SuperVP.
The results of analysis are stored in an SDIF file which pathname is returned.

- <infile> : pathname or SOUND object to be analysed

- <treshold> : transient detection treshold (1-10)

- <minfreq> : minimal detection frequency (Hz)

- <maxfreq< : maximal detection frequency (Hz)

- <minoffset> : min offset between markers (s)

- <windowsize> : number of samples of the analysis window

- <fftsize> : number of points of fft

- <step-oversamp> : oversampling proportion between two successive analysis windows

- <windowtype> : shape of the analysis window

- <out> : output file pathname
"
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let ((outname (if out 
                      (handle-new-file-exists (if (pathnamep out) out (outfile out)))
                      (om-choose-new-file-dialog :prompt "Choose a new SDIF markers file"
                                                 :directory (def-save-directory)))))
      (setf *last-saved-dir* (make-pathname :directory (pathname-directory outname)))
      (when outname
        (let* ((unix-outname (om-path2cmdpath outname))
               (fftstr (format nil "-N~D -M~D -W~D -oversamp ~D " fftsize windowsize windowtype step-oversamp))
               (transient-params (format nil "-OT -td_thresh ~D -td_G 2.5 -td_band ~D,~D -td_nument 10.0 -td_minoff ~Ds"
                                     treshold minfreq maxfreq minoffset))
               (cmd (format nil "~s -v -t -ns -S~s -Afft ~A ~A ~s" 
                            (om-path2cmdpath *SVP-PATH*)
                            (om-path2cmdpath infile)
                            fftstr transient-params
                            unix-outname))
               )
          (om-print "===========================")
          (om-print "SUPERVP TRANSIENT DETECTION")
          (om-print "===========================")
          (om-print cmd)   
          (om-cmd-line cmd *sys-console*)
          (and outname (probe-file outname)))))
    (om-beep-msg "!!! SuperVP not found !!!"))
  )


(defmethod! transient-detection ((infile pathName) &key
                         (treshold 1.4) (minfreq 0.0) (maxfreq 22050.0) (minoffset 0.02)
                         (windowsize 4096) (fftsize 4096) (step-oversamp 8) (windowtype "hanning") 
                          (out "markers.sdif"))
  (transient-detection (om-namestring infile) 
               :treshold treshold :minfreq minfreq :maxfreq maxfreq :minoffset minoffset
               :windowsize windowsize :fftsize fftsize :step-oversamp step-oversamp :windowtype windowtype 
               :out out))

(defmethod! transient-detection ((infile sound) &key 
                         (treshold 1.4) (minfreq 0.0) (maxfreq 22050.0) (minoffset 0.02)
                         (windowsize 4096) (fftsize 4096) (step-oversamp 8) (windowtype "hanning") 
                         (out "markers.sdif"))
  (transient-detection (sound-path infile) 
                       :treshold treshold :minfreq minfreq :maxfreq maxfreq :minoffset minoffset
                       :windowsize windowsize :fftsize fftsize :step-oversamp step-oversamp :windowtype windowtype 
                       :out out))  


;;;================================================================================================================
;;; FORMANT ANALYSIS
;;;================================================================================================================

(defmethod! formant-analysis ((infile string) &key
                              (analysis-mode 'lpc) (max-n-formants 45) (analysis-order 45) (extend-across-sound-limits nil)
                              (windowsize 2048) (fftsize 2048) (step-oversamp 8) (windowtype "blackman") 
                              (out "formants.sdif"))
  :initvals '(nil lpc 45 45 nil 2048 2048 8 "blackman" "formants.sdif")
  :menuins '((1 (("LPC" lpc) ("Discrete Spectrum" ds)))
             (4 (("No" nil) ("Yes" t)))
             (7 (("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
             (8 (("Blackman" "blackman") ("Hanning" "hanning") ("Hamming" "hamming"))))
  :icon 952
  :doc "Performs formant analysis of a sound using SuperVP.
The results of analysis are stored in an SDIF file which pathname is returned.

- <infile> the pathname or SOUND object to be analysed

- <analysis-mode> : Type of analysis (LPC or Discrete Spectrum)

- <analysis-order> : order of analysis

- <extend-across-sound-limits> determines whether the first analysis window is centered over the first sample (T) or starts at the first sample (NIL).

- <windowsize> : number of samples of the analysis window

- <fftsize> : number of points of fft

- <step-oversamp> : oversampling proportion between two successive analysis windows

- <windowtype> : shape of the analysis window

- <out> : output file pathname

"
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
      (let ((outname (if out 
                         (handle-new-file-exists (if (pathnamep out) out (outfile out)))
                      (om-choose-new-file-dialog :prompt "Choose a new SDIF file"
                                                 :directory (def-save-directory)))))
      (setf *last-saved-dir* (make-pathname :directory (pathname-directory outname)))
      (when outname
        (let* ((unix-outname (namestring outname))
               (fftstr (format nil "-N~D -M~D -W~D -oversamp ~D" fftsize windowsize windowtype step-oversamp))
               (formantsstr (string+ 
                             (if (equal analysis-mode 'ds) "-Aformant_ced" "-Aformant_lpc")
                             " n" (number-to-string max-n-formants) " " (number-to-string analysis-order)))
               (uflagstr (if extend-across-sound-limits "" "-U"))
               (cmd (format nil "~s -v -t -ns ~A -S~s ~A ~A -OS0 ~s" 
                            (namestring *SVP-PATH*)
                            uflagstr
                            (namestring infile)
                            fftstr formantsstr
                            unix-outname))
               )
          (om-print "==========================")
          (om-print "SUPERVP FORMANT ANALYSIS")
          (om-print "==========================")
          (om-print cmd)   
          (om-cmd-line cmd *sys-console*)
          (and outname (probe-file outname)))))
    (om-beep-msg "!!! SuperVP not found !!!"))
  )


(defmethod! formant-analysis ((infile pathname) &key
                              (analysis-mode 'lpc) (max-n-formants 45) (analysis-order 45) (extend-across-sound-limits nil)
                              (windowsize 2048) (fftsize 2048) (step-oversamp 8) (windowtype "blackman") 
                              (out "formants.sdif"))
  (formant-analysis (namestring infile) 
               :analysis-mode analysis-mode :max-n-formants max-n-formants :analysis-order analysis-order 
               :extend-across-sound-limits extend-across-sound-limits 
               :windowsize windowsize :fftsize fftsize :step-oversamp step-oversamp :windowtype windowtype 
               :out out))

(defmethod! formant-analysis ((infile sound) &key 
                              (analysis-mode 'lpc) (max-n-formants 45) (analysis-order 45) (extend-across-sound-limits nil)
                              (windowsize 2048) (fftsize 2048) (step-oversamp 8) (windowtype "blackman") 
                              (out "formants.sdif"))
            (formant-analysis (sound-path infile) 
                              :analysis-mode analysis-mode :max-n-formants max-n-formants :analysis-order analysis-order 
                              :extend-across-sound-limits extend-across-sound-limits 
                              :windowsize windowsize :fftsize fftsize :step-oversamp step-oversamp :windowtype windowtype 
                              :out out))  

