;;;===================================================
;;;
;;; OM-SuperVP
;;; SuperVP sound analysis and processing for OpenMusic
;;;
;;; Requires perliminary installation of SuperVP (also included in AudioSculpt)
;;; Set SuperVP path in OM preferences (externals) once the library is loaded.
;;;
;;;
;;; SPLIT/MERGE MULTICHANNEL FILES
;;; Authors: Jean Bresson, Marco Stroppa 
;;;
;;;===================================================


(in-package :om)


(defmethod! supervp-merge (files &optional outpath)
            :initvals '(nil)
            :indoc '("a list of sounds (or sound file pathnames) or a directory pathname" "output pathname")
            :icon 950
            :doc "
Merges all files in <files> into a single multi-channel audio file.

If <outname> is not specified, the merged file will be created in the in <files> if <files> is a directory, or in the OUT-FILES folder (see OM preferences) if not.
"
      (let ((file-list 
             (mapcar 'namestring 
                     (print (cond ((listp files) 
                            (remove nil (loop for item in files collect 
                                              (if (subtypep (type-of item) 'sound) 
                                                  (om-sound-file-name item)
                                                item))))
                           ((directoryp files)
                            (om-directory files :directories nil))
                           (t nil)))))
            (outname (namestring 
                      (cond ((pathnamep outpath) 
                             outpath)
                            ((and (directoryp files) (stringp outpath))
                             (om-make-pathname :directory files :name outpath))
                            ((directoryp files) 
                             (om-make-pathname :directory files :name "merge-file"))
                            (t (outfile "merge-file")))))
            (param-file (paramfile "tmp-merge.par"))
            )

        (with-open-file (out param-file :direction :output 
                             :if-does-not-exist :create :if-exists :supersede)
          (loop for elt in (cdr file-list) 
                for i = 2 then (+ i 1) do
                (format out "~s~%[~D/0,1]~%" elt i))
          )
        
        (supervp-command (format nil "-S~s -mixlist ~s ~s" (car file-list) (namestring param-file) outname))

        (when *delete-inter-file* (delete-file param-file))
        
        (or (probe-file outname)
            (om-beep-msg "SUPERVP merging error"))))



;;;======================================================================================

(defmethod sound-n-channels ((self sound))
  (om-sound-n-channels self))

(defmethod sound-n-channels ((self pathname))
  (let ((thesound (om-make-sound 'sound self))) 
    (om-sound-n-channels thesound)))

(defmethod sound-n-channels ((self string))
  (sound-n-channels (pathname self)))

(defmethod supervp-extract-one-channel (file n outfile)
  (supervp-command (format nil "-S~s -C~D ~s" (namestring file) n (namestring outfile)))
  (probe-file outfile))

(defmethod supervp-extract-one-channel ((file sound) n outfile)
  (supervp-extract-one-channel (om-sound-file-name file) n outfile))


(defmethod! supervp-split (file &optional channels outfolder)
            :initvals '(nil nil nil)
            :indoc '("a sound (or sound file pathname)" "channels to extract" "output directory pathname")
            :icon 950
            :doc "
Splits <file> into several mono sound files stored in <outfolder>.

<channels> can be an integer number or a list of numbers specifying specific channels to extract. If unspecific, all channels in <file> are extracted.
Channel files are named after the original <file> name cancatenated with the channel number (e.g. myfile-1.aiff, myfile-2.aiff, etc.)

If <outfolder> is not specified, the extracted channel files are stored in a sub-folder named after <file>.

"
      (let* ((nc (sound-n-channels file))
             (chans (if (null channels) (arithm-ser 1 nc 1)
                      (remove nil (loop for chan in (list! channels) collect 
                                        (if (and (integerp chan) (<= chan nc)) chan 
                                          (om-beep-msg (format nil "There is no channel ~D in ~s !" chan file)))))))
             (outfolder (or outfolder (om-make-pathname :directory (append (pathname-directory file) (list (pathname-name file)))))))
        
        (unless (probe-file outfolder)
          (om-create-directory outfolder))
        
        (if (and (integerp channels) (car chans))
            (supervp-extract-one-channel file (car chans)
                                           (om-make-pathname :directory (pathname-directory outfolder) 
                                                             :name (format nil "~a-~D" (pathname-name file) (car chans))
                                                             :type (pathname-type file)))
        (loop for channel in chans collect
              (supervp-extract-one-channel file channel
                                           (om-make-pathname :directory (pathname-directory outfolder) 
                                                             :name (format nil "~a-~D" (pathname-name file) channel)
                                                             :type (pathname-type file)))
              )
        )))


(defmethod! supervp-split ((file sound) &optional channels outfolder)
     (supervp-split (om-sound-file-name file) channels outfolder))      
              
        
        