;;;===================================================
;;;
;;; OM-SuperVP
;;; SuperVP sound analysis and processing for OpenMusic
;;;
;;; Requires perliminary installation of SuperVP (also included in AudioSculpt)
;;; Set SuperVP path in OM preferences (externals) once the library is loaded.
;;;
;;;
;;; SUPERVP PREFERENCES
;;; Integrated in the 'Externals' tab in OM preferences
;;; Authors: Jean Bresson (IRCAM - 2006-2010)
;;;
;;;===================================================



(in-package :om)


(defvar *SVP-PATH* "path to svp")
(defvar *param-files-folder* nil)

(defvar *delete-param-files* t)


;;;== PARAM FILES ===

(defmethod! set-paramfiles-folder ((path string))
  :icon 186
  (setf *param-files-folder* (pathname path)))

(defmethod! set-paramfiles-folder ((path pathname))
  :icon 186
  (setf *param-files-folder* path))

(defmethod! paramfile ((name string) &optional (unix nil))
  :icon 186
  (let ((pa (make-pathname :directory (pathname-directory *param-files-folder*) :name name
                           :host (pathname-host *param-files-folder*) 
                           :device (pathname-device *param-files-folder*))))
    (if unix (om-path2cmdpath pa) pa)))

(defmethod! paramfile ((name null) &optional (unix nil))
  :icon 186
  (let ((pa *param-files-folder*))
    (if unix (om-path2cmdpath pa) pa)))


;; (AddGenFun2Pack '(paramfile) *soundfolderspackage*)

;;;=== SVP PATH ===

(defmethod! set-svp-path ((path string))
  :icon 186
  (setf *SVP-PATH* (pathname path)))

(defmethod! set-svp-path ((path pathname))
  :icon 186
  (setf *SVP-PATH* path))

(defmethod! svp-path (&optional (unix nil))
  :icon 186
  (let ((pa *SVP-PATH*))
    (if unix (om-path2cmdpath pa) 
        pa)))


(defun init-svp ()
  (setf *SVP-PATH* (om-external-app '("AudioSculpt" "Kernels") "supervp"))
  (setf *param-files-folder* *om-infiles-folder*)
  )

;(add-init-user-func 'init-svp)    
(init-svp)


;;;=====================
;;; SVP PREFS

;(pushr 'svp *external-prefs*)
(add-external-pref-module 'svp)




(defmethod get-external-name ((module (eql 'svp))) "SuperVP")
(defmethod get-external-icon ((module (eql 'svp))) (and (exist-lib-p "OM-SuperVP") (list 950 (find-library "OM-SuperVP"))))

(defmethod get-external-module-vals ((module (eql 'svp)) modulepref) (get-pref modulepref :svp-options))
(defmethod get-external-module-path ((module (eql 'svp)) modulepref) (get-pref modulepref :svp-path))
(defmethod set-external-module-vals ((module (eql 'svp)) modulepref vals) (set-pref modulepref :svp-options vals))
(defmethod set-external-module-path ((module (eql 'svp)) modulepref path) 
  (set-pref modulepref :svp-path path))


;;; params folder
(defun def-svp-options ()
  (list *om-infiles-folder*))

(defmethod get-external-def-vals ((module (eql 'svp))) 
  (list :svp-path (om-external-app '("AudioSculpt" "Kernels") "supervp")
        :svp-options (def-svp-options)))

(defmethod save-external-prefs ((module (eql 'svp))) 
  `(:svp-path ,(om-save-pathname *SVP-PATH*) 
    :svp-options (list ,(om-save-pathname *param-files-folder*))))


(defmethod put-external-preferences ((module (eql 'svp)) moduleprefs)
  (let ((list-prefs (get-pref moduleprefs :svp-options)))
    (when list-prefs 
      (if (probe-file (nth 0 list-prefs))
       (setf *param-files-folder* (nth 0 list-prefs))
       (progn
         (setf (nth 0 list-prefs) (nth 0 
                                       (nth (+ 1 (position :svp-options (get-external-def-vals module)))
                                            (get-external-def-vals module))))
         (set-pref moduleprefs :svp-options list-prefs)
         ;(push :svp-params-dir *restore-defaults*)
         ))
      )
    (when (get-pref moduleprefs :svp-path)
      (setf *SVP-PATH* (find-true-external (get-pref moduleprefs :svp-path))))
    ))

(put-external-pref-values 'svp)

(defmethod show-external-prefs-dialog ((module (eql 'svp)) prefvals)
  (let* ((rep-list (copy-list prefvals))
         (dialog (om-make-window 'om-dialog
                                 :window-title "SuperVP Options"
                                 :size (om-make-point 300 160)
                                 :position :centered
                                 :resizable nil :maximize nil :close nil))
         paramstext)
    (om-add-subviews dialog
                     (om-make-dialog-item 'om-static-text  (om-make-point 20 20) (om-make-point 200 20) 
                                          "Param Files folder (read/write):" :font *controls-font*)
                     (setf paramstext (om-make-dialog-item 'om-static-text  (om-make-point 20 45) (om-make-point 220 50) 
                                                           (om-namestring (nth 0 prefvals))
                                                           :font *om-default-font1*))
                     (om-make-dialog-item 'om-button (om-make-point 240 44) (om-make-point 50 24) "..." 
                                          :di-action (om-dialog-item-act item
                                                       (declare (ignore item))
                                                       (let ((newpath (om-choose-directory-dialog 
                                                                       :directory (om-namestring (nth 0 prefvals)))))
                                                         (when newpath 
                                                           (om-set-dialog-item-text paramstext (om-namestring newpath))
                                                           ))))
                     ;;; boutons
                     (om-make-dialog-item 'om-button (om-make-point 20 110) (om-make-point 80 20) "Restore"
                                          :di-action (om-dialog-item-act item
                                                       (om-set-dialog-item-text paramstext (namestring (nth 0 (def-svp-options))))
                                                       ))
      
      (om-make-dialog-item 'om-button (om-make-point 130 110) (om-make-point 80 20) "Cancel"
                           :di-action (om-dialog-item-act item
                                        (om-return-from-modal-dialog dialog nil)))
      
      (om-make-dialog-item 'om-button (om-make-point 210 110) (om-make-point 80 20) "OK"
                           :di-action (om-dialog-item-act item
                                        (let ((argerror nil)
                                              (partxt (om-dialog-item-text paramstext)))
                                          (setf (nth 0 rep-list) partxt)
                                          (if argerror
                                              (om-message-dialog (format nil "Error in a SuperVP option.~% Preferences could not be recorded.")) 
                                            (om-return-from-modal-dialog dialog rep-list))
                                          ))
                           :default-button t :focus t)
                     
      )
    (om-modal-dialog dialog)))





