;;;===================================================
;;;
;;; OM-SuperVP
;;; SuperVP sound analysis and processing for OpenMusic
;;;
;;; Requires perliminary installation of SuperVP (also included in AudioSculpt)
;;; Set SuperVP path in OM preferences (externals) once the library is loaded.
;;;
;;;
;;; SOUND PROCESSING
;;; Transposition, Pitch shifting, Time stretching, Band filter, Format filter, 
;;; Surface filter, Break-point filter, Clipping, Freezing, Gain ...
;;;
;;; Authors: Jean Bresson & Jean Lochard (IRCAM - 2006-2010)
;;;
;;;===================================================

(in-package :om)
  

(defvar *tmpparfiles* nil)

(defmethod make-temp-param ((self bpf) &optional (filename "svp-tmp-param.par"))
  (let ((tmpfile (om::paramfile filename)))
    (om::save-params self tmpfile)
    (push tmpfile *tmpparfiles*)
    (om::om-path2cmdpath tmpfile)))

(defmethod make-temp-param ((self textfile) &optional (filename "svp-tmp-param.par"))
  (let ((tmpfile (om::paramfile filename)))
    (om::save-params self tmpfile)
    (push tmpfile *tmpparfiles*)
    (om::om-path2cmdpath tmpfile)))

(defmethod make-temp-param ((self list) &optional (filename "svp-tmp-param.par"))
  (let ((tmpfile (om::paramfile filename)))
    (om::save-params self tmpfile)
    (push tmpfile *tmpparfiles*)
    (om::om-path2cmdpath tmpfile)))

(defmethod make-temp-param ((self pathname) &optional (filename "svp-tmp-param.par"))
  (om::om-path2cmdpath self))

(defmethod make-temp-param ((self string) &optional (filename "svp-tmp-param.par"))
  (om::om-path2cmdpath (if (probe-file (pathname self)) self (om::paramfile self))))

(defmethod make-temp-param ((self number) &optional (filename "svp-tmp-param.par"))
  (format nil "~D" self))





;;; FORMAT SVP COMMAND
(om::defmethod! make-svp-command (srcpath processings begin end
                                      windowsize fftsize windowstep-oversamp
                                      window-type shape-invariant preserve-transient
                                      normalize
                                      outpath)
  :icon 950 
  :menuins '((4 (("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
             (5 (("Blackman" "blackman") ("Hanning" "hanning")("Hamming" "hamming")))
             (6 (("Shape Invariant On" t) ("Shape Invariant Off" nil)))
             (8 (("Normalize On" t) ("Normalize Off" nil)))
             )
  :initvals '(nil "" 4096 4096 8 "hanning" nil t nil "out.aiff")
  (if (and *SVP-PATH* (probe-file *SVP-PATH*))
    (let ((srcpath (pathname srcpath))
          (beginstr (if begin (format nil "-B~D " begin) "")) 
          (endstr (if end (format nil "-E~D " end) ""))
          (command ""))
      (when (probe-file srcpath)
        (setf command (om::string+ command (format nil "~s -t -Z " 
                                               (om::om-path2cmdpath *SVP-PATH*)
                                               )
                               beginstr endstr))
        
        (when normalize
          (setf command (om::string+ command "-norm ")))
        
        (setf command (om::string+ command (format nil "-U -S~s -Afft " 
                                               (om::om-path2cmdpath srcpath)
                                               )))

        (setf command (om::string+ command (format nil "-M~D -N~D -W~D "
                                              windowsize fftsize  window-type)
                                   (if windowstep-oversamp (format nil "-oversamp ~D " windowstep-oversamp ""))))
        (when shape-invariant
          (setf command (om::string+ command "-shape 2 -Vuf -4 ")))
        
        ;;; transients
        (if preserve-transient
          (let ((tr-settings (when preserve-transient
                         (if (listp preserve-transient) preserve-transient
                             '(1.4 (0.0 22050.0) 1.5 nil)))))

            (setf command (om::string+ command (format nil "-P1 -td_thresh ~D -td_band ~D,~D -td_ampfac ~D "
                                                   (nth 0 tr-settings)
                                                   (car (nth 1 tr-settings))
                                                   (cadr (nth 1 tr-settings))
                                                   (nth 2 tr-settings)
                                                   )))
          (when (nth 4 tr-settings)
            (setf command (om::string+ command (format nil "-td_relaxto ~D -td_relax ~D ")
                                   (car (nth 3 tr-settings))
                                   (cadr (nth 3 tr-settings))
                                   )))
          )
          (unless (find "-Ffshift" (om::list! processings) :test #'(lambda (s1 s2) 
                                                                     (and (>= (length s2) 8)  
                                                                          (string-equal s1 s2 :end1 8 :end2 8))))
            ;;; option -P0 (phase synchronization) incompatible avec freq shifting
            (setf command (om::string+ command (format nil "-P0 "))))
          )
        
        ;;; processings
        (loop for p in (om::list! processings) do
              (setf command (om::string+ command p " ")))
        
        ;;; out
        (setf command (om::string+ command (format nil "-Osa~D ~s" *audio-res* (om::om-path2cmdpath outpath))))
        
      command))
  
    (om-beep-msg "SuperVP not found.")
    ))

(om::defmethod! supervp-processing ((srcpath pathname) processings begin end windowsize fftsize windowstep-oversamp 
                                     window-type shape-invariant preserve-transient
                                     normalize outfile)
  :icon 950 
  :menuins '((6 (("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
             (7 (("Blackman" "blackman") ("Hanning" "hanning")("Hamming" "hamming")))
             (8 (("Shape Invariant On" t) ("Shape Invariant Off" nil)))
             (10 (("Normalize On" t) ("Normalize Off" nil))))
  :initvals '(nil "" nil nil 4096 4096 8 "hanning" nil t nil "out.aiff")
  (let* ((outpath (if (pathnamep outfile) outfile (om::outfile outfile)))
         (cmd (make-svp-command srcpath processings begin end windowsize fftsize windowstep-oversamp 
                                window-type shape-invariant preserve-transient normalize
                                outpath)))
    (when cmd
      (om-print "==========================")
      (om-print "SUPERVP PROCESSING")
      (om-print "==========================")
      
      (om-print (format nil "Input file: ~s" (namestring srcpath)))
      (om-print (format nil "Output file: ~s" (namestring outpath)))
     
      (om::om-cmd-line cmd om::*sys-console*)
      (when *delete-inter-file* (loop for file in *tmpparfiles* do (delete-file file)))
      (setf *tmpparfiles* nil)
      (probe-file outpath))))
    
(om::defmethod! supervp-processing ((srcpath string) processings begin end windowsize fftsize windowstep-oversamp 
                                     window-type shape-invariant preserve-transient
                                     normalize outfile)
  (let ((file (if (probe-file (pathname srcpath)) 
                   (pathname srcpath)
                 (infile srcpath))))
    (supervp-processing file 
                        processings begin end windowsize fftsize windowstep-oversamp 
                        window-type shape-invariant preserve-transient normalize
                        outfile)))
         
(om::defmethod! supervp-processing ((srcpath sound) processings begin end windowsize fftsize windowstep-oversamp 
                                     window-type shape-invariant preserve-transient
                                     normalize outfile)
         (supervp-processing (om-sound-file-name srcpath) 
                             processings begin end windowsize fftsize windowstep-oversamp 
                             window-type shape-invariant preserve-transient normalize
                             outfile))





;;;=================================================================================================
;;; TIME STRETCH
;;;=================================================================================================

(om::defmethod! supervp-timestretch ((self number))
    :icon 951 
    (format nil "-D~D" self))

(om::defmethod! supervp-timestretch ((self pathname))
    :icon 951 
    (format nil "-D~s" (om::om-path2cmdpath self)))

(om::defmethod! supervp-timestretch ((self string))
    :icon 951 
    (format nil "-D~s" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))))

(om::defmethod! supervp-timestretch ((self om::textfile))
    :icon 951 
    (let ((tmpfile (om::paramfile "temptimestretch.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-D\"~a\"" (namestring tmpfile))))

(om::defmethod! supervp-timestretch ((self om::bpf))
    :icon 951 
    (let ((tmpfile (om::paramfile "temptimestretch.par")))
     (om::save-params self tmpfile)
     (push tmpfile *tmpparfiles*)
     (format nil "-D\"~a\"" (namestring tmpfile))))

(om::defmethod! supervp-timestretch ((self list))
    :icon 951 
    (let ((tmpfile (om::paramfile "temptimestretch.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-D\"~a\"" (namestring tmpfile))))


;;;=================================================================================================
;;; TRANSPOSITION
;;;=================================================================================================

;;; GET TRANS PARAMS 
;;; returns number or list ((t1 val1) (t2 val2) ...)
(defmethod get-trans-params ((self number)) self)

(defmethod get-trans-params ((self list)) self)

(defmethod get-trans-params ((self bpf)) 
  (mat-trans (list (x-points self) (y-points self))))

(defmethod get-trans-params ((self textfile)) 
  (list-of-data (buffer-text self)))

(defmethod get-trans-params ((self string)) (get-trans-params (pathname self)))

(defmethod get-trans-params ((self pathname)) 
  (when (probe-file self)
    (let ((tmpbuffer (om-make-buffer))
          (lst nil))
      (om-buffer-insert-file tmpbuffer self)
      (setf lst (list-of-data tmpbuffer))
      (om-kill-buffer tmpbuffer)
      lst)))

;;; renvoie deux listes echantillonnes aux mmes instants...
(defun interpole-envelopes (env1 env2)
  (let ((env1x (mapcar 'car env1))
        (env1y (mapcar 'cadr env1))
        (env2x (mapcar 'car env2))
        (env2y (mapcar 'cadr env2))
        newenv1pts newenv2pts)
  (loop for element in env1 do
        (unless (member (car element) env2 :test '= :key 'car)
          (push (list (car element) (table_xy env2x env2y (car element))) newenv2pts)
          ))
  (loop for element in env2 do
        (unless (member (car element) env1 :test '= :key 'car)
          (push (list (car element) (table_xy env1x env1y (car element))) newenv1pts)
          ))
  (values 
   (sort (append env1 newenv1pts) '< :key 'car)
   (sort (append env2 newenv2pts) '< :key 'car)
   )))


;;; transposition + envelope
(defmethod trans-params-2 ((trans number) (env number) trstr)
  (format nil "-transke ~D ~A ~D" (- trans env) trstr env))

(defmethod trans-params-2 ((trans list) (env number) trstr)
  (let ((trlist (mapcar #'(lambda (elem) (list (car elem) (- (cadr elem) env))) trans)))
    (format nil "-transke \"~a\" ~A ~D" (make-temp-param trlist "temp-transp.par") trstr env)
    ))
    
(defmethod trans-params-2 ((trans number) (env list) trstr)
  (let ((trlist (mapcar #'(lambda (elem) (list (car elem) (- trans (cadr elem)))) env)))
    (format nil "-transke \"~a\" ~A \"~a\"" (make-temp-param trlist "temp-transp.par") trstr (make-temp-param env "temp-env-transp.par"))
    ))

(defmethod trans-params-2 ((trans list) (env list) trstr)
  
    (multiple-value-bind (itrans ienv)
        (interpole-envelopes trans env)
      (let ((trlist (mapcar #'(lambda (tr en) (list (car tr) (- (cadr tr) (cadr en)))) itrans ienv)))
        (format nil "-transke \"~a\" ~A \"~a\"" (make-temp-param trlist "temp-transp.par") trstr (make-temp-param env "temp-env-transp.par")))
      ))

;;; transposition seule
(defun trans-params-1 (param trstr)
  (string+ trstr " " (make-temp-param param "temptransp.par")))

;;; full command line
(defun transp-cmd (param preserve env order envtrans timecor mode)
  (let ((transstring (if timecor "-trans" "-transnc")))
    (if (null preserve)
        (om::string+ " " (trans-params-1 param transstring) 
                     " -Afft" 
                     " -transmo " (om::integer-to-string mode))
      (om::string+ (trans-params-2 (get-trans-params param) (get-trans-params envtrans) transstring) 
                   " -Afft " (om::integer-to-string order) (if env "t" "")
                   " -transmo " (om::integer-to-string mode)))
    ))
 
;;; OM Methods 
(om::defmethod! supervp-transposition ((self number) preserve-enveloppe enveloppe-type filter-order &optional (envtransp 0) (time-correction t) (mode 0))
    :icon 951 
    :menuins '((1 (("Preserve Enveloppe On" t) ("Preserve Enveloppe Off" nil)))
               (2 (("True Enveloppe" t) ("LPC Enveloppe" nil)))
               (5 (("Time correction On" t) ("Time correction Off" nil)))
               (6 (("Time Domain" 0) ("Automatic" 1) ("Frequency Domain" 2))))
    :initvals '(0 t t 70 0 t 0)
   (transp-cmd self preserve-enveloppe enveloppe-type filter-order envtransp time-correction mode))

(om::defmethod! supervp-transposition ((self pathname) preserve-enveloppe enveloppe-type filter-order &optional (envtransp 0) (time-correction t) (mode 0))
    (transp-cmd self preserve-enveloppe enveloppe-type filter-order envtransp time-correction mode))

(om::defmethod! supervp-transposition ((self string) preserve-enveloppe enveloppe-type filter-order &optional (envtransp 0) (time-correction t) (mode 0))
    (transp-cmd self preserve-enveloppe enveloppe-type filter-order envtransp time-correction mode))

(om::defmethod! supervp-transposition ((self om::textfile) preserve-enveloppe enveloppe-type filter-order &optional (envtransp 0) (time-correction t) (mode 0))
      (transp-cmd self preserve-enveloppe enveloppe-type filter-order envtransp time-correction mode))

(om::defmethod! supervp-transposition ((self om::bpf) preserve-enveloppe enveloppe-type filter-order &optional (envtransp 0) (time-correction t) (mode 0))
      (transp-cmd self preserve-enveloppe enveloppe-type filter-order envtransp time-correction mode))

(om::defmethod! supervp-transposition ((self list) preserve-enveloppe enveloppe-type filter-order &optional (envtransp 0) (time-correction t) (mode 0))
    (transp-cmd self preserve-enveloppe enveloppe-type filter-order envtransp time-correction mode))


;;;=================================================================================================
;;; FREQUENCY-SHIFTING
;;;=================================================================================================


(om::defmethod! supervp-frequencyshift ((self number))
    :icon 951 
    (let ((tmpfile (om::paramfile "fshift.par")))
      (om::save-params (list (list 0 self)) tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Ffshift \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-frequencyshift ((self pathname))
    :icon 951 
    (format nil "-Ffshift \"~a\"" (om::om-path2cmdpath self)))

(om::defmethod! supervp-frequencyshift ((self string))
    :icon 951 
    (format nil "-Ffshift \"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))))

(om::defmethod! supervp-frequencyshift ((self om::textfile))
    :icon 951 
    (let ((tmpfile (om::paramfile "fshift.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Ffshift \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-frequencyshift ((self om::bpf))
    :icon 951 
    (let ((tmpfile (om::paramfile "fshift.par")))
     (om::save-params self tmpfile)
     (push tmpfile *tmpparfiles*)
     (format nil "-Ffshift \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-frequencyshift ((self list))
    :icon 951 
    (let ((tmpfile (om::paramfile "fshift.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Ffshift \"~a\"" (om::om-path2cmdpath tmpfile))))

;;;=================================================================================================
;;; BREAKPOINT-FILTER
;;;=================================================================================================

(defun breakpt-cmd (param extrapol)
  (if extrapol
    (om::string+ "-Fbreakpt " param)
    (om::string+ "-Fbreakpt-noex " param)))


(om::defmethod! supervp-breakpointfilter  ((self list) extrapolation)
    :icon 951
    :menuins '((1 (("Extrapolation On" t) ("Extrapolation Off" nil))))
    :initvals '(((0.1 0 4 100 0 500 -30 1000 -30 4000 0)) t) 
    (let ((tmpfile (om::paramfile "breakpt.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (breakpt-cmd (format nil "~s" (om::om-path2cmdpath tmpfile)) extrapolation)))

(om::defmethod! supervp-breakpointfilter   ((self pathname) extrapolation)
    :icon 951 
    (fbreakpt-cmd (format nil "\"~a\"" (om::om-path2cmdpath self)) extrapolation))

(om::defmethod! supervp-breakpointfilter  ((self string) extrapolation)
    :icon 951 
    (breakpt-cmd (format nil "\"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))) extrapolation))

(om::defmethod! supervp-breakpointfilter   ((self om::textfile) extrapolation)
    :icon 951 
    (let ((tmpfile (om::paramfile "breakpt.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (breakpt-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) extrapolation)))


;;;=================================================================================================
;;; FORMANT FILTER 
;;;=================================================================================================


(defun formant-cmd (param interpol extrapol)
  (if interpol
    (om::string+ (if extrapol "-Ffof " "-Ffof-noex ") param)
    (om::string+ (if extrapol "-Ffifof " "-Ffifof-noex ") param)))
    
(om::defmethod! supervp-formantfilter ((self list) interpolation extrapolation)
    :icon 951 
    :menuins '((1 (("Interpolation On" t) ("Interpolation Off" nil)))
               (2 (("Extrapolation On" t) ("Extrapolation Off" nil))))
    :initvals '(((0 2 350 0 50 1700 -5 100)) t t)

    (let ((tmpfile (om::paramfile "formant.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (formant-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) interpolation extrapolation)))
    

(om::defmethod! supervp-formantfilter ((self pathname) interpolation extrapolation)
    :icon 951 
    (formant-cmd (format nil "\"~a\"" (om::om-path2cmdpath self)) interpolation extrapolation))

  
(om::defmethod! supervp-formantfilter ((self string) interpolation extrapolation)
    :icon 951 
    (formant-cmd (format nil "\"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))) interpolation extrapolation))

(om::defmethod! supervp-formantfilter ((self om::textfile) interpolation extrapolation)
    :icon 951 
    (let ((tmpfile (om::paramfile "formant.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (formant-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) interpolation extrapolation)))

(om::defmethod! supervp-formantfilter ((self om::bpf) interpolation extrapolation)
    :icon 951 
    (let ((tmpfile (om::paramfile "formant.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (formant-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) interpolation extrapolation)))


;;;=================================================================================================
;;; BAND-FILTER
;;;=================================================================================================

(defun fband-cmd (param extrapol)
  (if extrapol
    (om::string+ "-Fbande " param)
    (om::string+ "-Fbande-noex " param)))


(om::defmethod! supervp-bandfilter  ((self list) extrapolation)
    :icon 951
    :menuins '((1 (("Extrapolation On" t) ("Extrapolation Off" nil))))
    :initvals '(((0 2 350 500)) t) 
    (let ((tmpfile (om::paramfile "fband.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (fband-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) extrapolation)))

(om::defmethod! supervp-bandfilter  ((self pathname) extrapolation)
    :icon 951 
    (fband-cmd (format nil "\"~a\"" (om::om-path2cmdpath self)) extrapolation))

(om::defmethod! supervp-bandfilter ((self string) extrapolation)
    :icon 951 
    (fband-cmd (format nil "\"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))) extrapolation))

(om::defmethod! supervp-bandfilter  ((self om::textfile) extrapolation)
    :icon 951 
    (let ((tmpfile (om::paramfile "fband.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (fband-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) extrapolation)))


;;;=================================================================================================
;;; CLIPING FILTER 
;;;=================================================================================================

(defun clipping-cmd (param renorm)
  (if renorm
    (om::string+ "-Fclip-norm " param)
    (om::string+ "-Fclip " param)))


(om::defmethod! supervp-clipping ((self pathname) renormalize)
    :icon 951
    :menuins '((1 (("Renormalize On" t) ("Renormalize Off" nil))))
    :initvals '(((-70 -60)) nil)
 
    (clipping-cmd (format nil "\"~a\"" (om::om-path2cmdpath self)) renormalize))

(om::defmethod! supervp-clipping ((self string) renormalize)
    :icon 951 
    (clipping-cmd (format nil "\"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))) renormalize))

(om::defmethod! supervp-clipping ((self om::textfile) renormalize)
    :icon 951 
    (let ((tmpfile (om::paramfile "clipping.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (clipping-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) renormalize)))

(om::defmethod! supervp-clipping ((self list) renormalize)
    :icon 951 
    (let ((tmpfile (om::paramfile "clipping.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (clipping-cmd (format nil "\"~a\"" (om::om-path2cmdpath tmpfile)) renormalize)))


;;;=================================================================================================
;;; FREEZE (Using SuperVP newfreeze)
;;;=================================================================================================

(om::defmethod! supervp-freeze ((self pathname))
    :icon 951
    :initvals '(((0.1 0.4 0.1)) nil) 
    (format nil "-Anewfreeze \"~a\"" (om::om-path2cmdpath self)))

(om::defmethod! supervp-freeze ((self string))
    :icon 951 
    (format nil "-Anewfreeze \"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))))

(om::defmethod! supervp-freeze ((self om::textfile))
    :icon 951 
    (let ((tmpfile (om::paramfile "freeze.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Anewfreeze \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-freeze ((self om::bpf))
    :icon 951 
    (let ((tmpfile (om::paramfile "freeze.par")))
     (om::save-params self tmpfile)
     (push tmpfile *tmpparfiles*)
     (format nil "-Anewfreeze \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-freeze ((self list))
    :icon 951 
    (let ((tmpfile (om::paramfile "freeze.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Anewfreeze \"~a\"" (om::om-path2cmdpath tmpfile))))



;;;=================================================================================================
;;; SURFACE FILTER 
;;;=================================================================================================


(om::defmethod! supervp-surfacefilter ((self pathname))
    :icon 951 
    (format nil "-Fsurface \"~a\"" (om::om-path2cmdpath self)))

(om::defmethod! supervp-surfacefilter ((self string))
    :icon 951 
    (format nil "-Fsurface \"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))))

(om::defmethod! supervp-surfacefilter ((self om::textfile))
    :icon 951 
    (let ((tmpfile (om::paramfile "surfacefilter.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Fsurface \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-surfacefilter ((self om::bpf))
    :icon 951 
    (let ((tmpfile (om::paramfile "surfacefilter.par")))
     (om::save-params self tmpfile)
     (push tmpfile *tmpparfiles*)
     (format nil "-Fsurface \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-surfacefilter ((self list))
    :icon 951 
    (let ((tmpfile (om::paramfile "surfacefilter.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Fsurface \"~a\"" (om::om-path2cmdpath tmpfile))))

;;;=================================================================================================
;;; FILE-POS
;;;=================================================================================================

(om::defmethod! supervp-filepos  ((self pathname))
    :icon 951 
    (format nil "-Ipos \"~a\"" (om::om-path2cmdpath self)))

(om::defmethod! supervp-filepos  ((self string))
    :icon 951 
    (format nil "-Ipos \"~a\"" (om::om-path2cmdpath (if (probe-file (pathname self))
                  self (om::paramfile self)))))

(om::defmethod! supervp-filepos  ((self om::textfile))
    :icon 951 
    (let ((tmpfile (om::paramfile "posfile.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Ipos \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-filepos  ((self list))
    :icon 951 
    (let ((tmpfile (om::paramfile "posfile.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-Ipos \"~a\"" (om::om-path2cmdpath tmpfile))))


;;;=================================================================================================
;;; GAIN / AMPLITUDE ENVELOPE
;;;=================================================================================================


(om::defmethod! supervp-gain ((self number))
    :icon 951 
    (format nil "-ggain ~D" self))

(om::defmethod! supervp-gain ((self pathname))
    :icon 951 
    (format nil "-ggain \"~a\"" (om::om-path2cmdpath self)))

(om::defmethod! supervp-gain ((self string))
    :icon 951 
    (format nil "-ggain \"~a\"" (om::om-path2cmdpath 
                               (if (probe-file (pathname self))
                                   self (om::paramfile self)))))

(om::defmethod! supervp-gain ((self om::textfile))
    :icon 951 
    (let ((tmpfile (om::paramfile "gain.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-ggain \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-gain ((self om::bpf))
    :icon 951 
    (let ((tmpfile (om::paramfile "gain.par")))
     (om::save-params self tmpfile)
     (push tmpfile *tmpparfiles*)
     (format nil "-ggain \"~a\"" (om::om-path2cmdpath tmpfile))))

(om::defmethod! supervp-gain ((self list))
    :icon 951 
    (let ((tmpfile (om::paramfile "gain.par")))
      (om::save-params self tmpfile)
      (push tmpfile *tmpparfiles*)
      (format nil "-ggain \"~a\"" (om::om-path2cmdpath tmpfile))))



;;;=================================================================================================
;;; PACK
;;;=================================================================================================


;(defvar *svp-treatments-package* (omNG-protect-object (omNG-make-new-package "Treatments")))
;(defvar *svp-processing-package* (omNG-protect-object (omNG-make-new-package "Processing")))
;(addPackage2Pack *svp-treatments-package* *svp-package*)
;(addPackage2Pack *svp-processing-package* *svp-package*)
;(AddGenFun2Pack  '(supervp-transposition supervp-timestretch
;                   supervp-frequencyshift supervp-breakpointfilter
;                   supervp-formantfilter supervp-bandfilter
;                   supervp-clipping supervp-freeze
;                   supervp-surfacefilter) *svp-treatments-package*)
;(AddGenFun2Pack  '(supervp-processing) *svp-processing-package*)