;;;===================================================
;;;
;;; OM-SuperVP
;;; SuperVP sound analysis and processing for OpenMusic
;;;
;;; Requires perliminary installation of SuperVP (also included in AudioSculpt)
;;; Set SuperVP path in OM preferences (externals) once the library is loaded.
;;;
;;;
;;; CROSS SYNTHESIS & SOURCE/FILTER SYNTHESIS
;;; 
;;; Authors: Jean Bresson (IRCAM - 2006-2010)
;;;
;;;===================================================




(in-package :om)

;;;=================================================================================================
;;; SYNTHESE CROISEE SOURCE FILTRE
;;;=================================================================================================

(defmethod get-src-path ((self pathname)) self)
(defmethod get-src-path ((self om::sound)) (sound-path self))
(defmethod get-src-path ((self string)) (pathname self))

(om::defmethod! supervp-sourcefilter-synthesis ((source t) (filtre t)  windowsize fftsize windowstep-oversamp 
                                     window-type normalize crossmode spectenv-type outfile)
  :icon 950 
  :menuins '((4 (("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
             (5 (("Blackman" "blackman") ("Hanning" "hanning")("Hamming" "hamming")))
             (6 (("Normalize On" t) ("Normalize Off" nil)))
             (7 (("Amplitude+Phase" 0) ("Amplitude" 1)("Phase" 2)))
             (8 (("LPC" t) ("True Enveloppe" nil)))
             )
  :initvals '(nil nil 4096 4096 8 "hanning" nil 0 t "out.aiff")
  (if (and om::*SVP-PATH* (probe-file om::*SVP-PATH*))
    (let ((srcpath (get-src-path source))
          (srcpath2 (get-src-path filtre)))
      (if (and (probe-file srcpath) (probe-file srcpath2))
      (let ((outpath (if (pathnamep outfile) outfile (outfile outfile)))
            (command ""))
  
        (setf command (om::string+ command  (format nil "~s -t -v -Z -U -u " 
                                                (om::om-path2cmdpath om::*SVP-PATH*))))
        
        (when normalize
          (setf command (om::string+ command "-norm ")))
        
        
        (setf command (string+ command  (format nil "-S~s -s~s -A " 
                                                (om::om-path2cmdpath srcpath)
                                                (om::om-path2cmdpath srcpath2))))
        
        (if spectenv-type (setf command (string+ command "-alpc ")) (setf command (om::string+ command "-atenv ")))
        
        (case crossmode
          (0 (setf command (om::string+ command "-Gmul ")))
          (1 (setf command (om::string+ command "-Gamul ")))
          (2 (setf command (om::string+ command "-Gpmul "))))
        
        
        (setf command (om::string+ command  (format nil "-M~D -N~D -m~D -n~D " windowsize fftsize windowsize fftsize ) 
                                   (if windowstep-oversamp (format nil "-oversamp ~D " windowstep-oversamp) "")
                                   (format nil "-W~s  -w~s " window-type window-type)
                                   (format nil "~s " (om::om-path2cmdpath outpath)))) 
                                            
        (print command)
        (om::om-cmd-line command om::*sys-console*)
        (loop for file in *tmpparfiles* do (delete-file file))
        (setf *tmpparfiles* nil)
        outpath)
      (om::om-beep-msg "Source / filter sound files no found")
      ))
    (om-beep-msg "SUPERVP not found!"))
  )





;;;=================================================================================================
;;; CROSS SYNTHESIS
;;;=================================================================================================

(om::defmethod! supervp-cross-synthesis ((sound1 t) (sound2 t) (param pathname) windowsize fftsize windowstep-oversamp 
                                     window-type normalize crossmode outfile)
  :icon 950 
  :menuins '((5 (("1/4" 4) ("1/8" 8) ("1/16" 16) ("1/32" 32)))
             (6 (("Blackman" "blackman") ("Hanning" "hanning")("Hamming" "hamming")))
             (7 (("Normalize On" t) ("Normalize Off" nil)))
             (8 (("Cross" t) ("Add" nil)))
             )
  :initvals '(nil nil "" 4096 4096 8 "hanning" nil nil "out.aiff")
  (if (and om::*SVP-PATH* (probe-file om::*SVP-PATH*))
  (let ((srcpath (get-src-path sound1))
        (srcpath2 (get-src-path sound2)))
    (if (and (probe-file srcpath) (probe-file srcpath2))
      (let ((outpath (if (pathnamep outfile) outfile (outfile outfile)))
            (command ""))
        (setf command (om::string+ command  (format nil "~s -t -v -Z -U -u " 
                                                (om::om-path2cmdpath om::*SVP-PATH*))))
        
        (when normalize
          (setf command (om::string+ command "-norm ")))

        
        (setf command (om::string+ command  (format nil "-S~s -s~s -A -a "  
                                                (om::om-path2cmdpath srcpath)
                                                (om::om-path2cmdpath srcpath2))))
        
        (if crossmode
          (setf command (om::string+ command "-Gcross "))
          (setf command (om::string+ command "-Gadd ")))
        
        (setf command (om::string+ command  (format nil "~s -M~D -N~D -m~D -n~D " 
                                                    (om-path2cmdpath param)
                                                    windowsize fftsize windowsize fftsize)
                                   (if  windowstep-oversamp (format nil "-oversamp ~D " windowstep-oversamp) "")
                                   (format nil "-W~s -w~s " window-type window-type)
                                   (format nil "~s " (om::om-path2cmdpath outpath))))
        (print command)
        (om::om-cmd-line command om::*sys-console*)
        (loop for file in *tmpparfiles* do (delete-file file))
        (setf *tmpparfiles* nil)
        outpath)
      (om::om-beep-msg "sound files no found"))
    )
  (om-beep-msg "SUPERVP not found!")))


(om::defmethod! supervp-cross-synthesis ((sound1 t) (sound2 t) (param string) windowsize fftsize windowstep-oversamp 
                                     window-type normalize crossmode outfile)
  (let ((parampath (if (probe-file (pathname param)) (pathname param) (paramfile param))))
    (supervp-cross-synthesis sound1 sound2 parampath windowsize fftsize windowstep-oversamp 
                             window-type normalize crossmode outfile)))


(om::defmethod! supervp-cross-synthesis ((sound1 t) (sound2 t) (param om::bpf) windowsize fftsize windowstep-oversamp 
                                     window-type normalize crossmode outfile)
  (let ((parampath (om::paramfile "tempcross.par"))
        (rep nil))
    (om::save-params param parampath)
    (setf rep (supervp-cross-synthesis sound1 sound2 parampath windowsize fftsize windowstep-oversamp 
                                       window-type normalize crossmode outfile))
    (delete-file parampath)
    rep))


(om::defmethod! supervp-cross-synthesis ((sound1 t) (sound2 t) (param list) windowsize fftsize windowstep-oversamp 
                                     window-type normalize crossmode outfile)
  (let ((parampath (om::paramfile "tempcross.par"))
        (rep nil))
    (om::save-params param parampath)
    (setf rep (supervp-cross-synthesis sound1 sound2 parampath windowsize fftsize windowstep-oversamp 
                                       window-type normalize crossmode outfile))
    (delete-file parampath)
    rep))

(om::defmethod! supervp-cross-synthesis ((sound1 t) (sound2 t) (param om::textfile) windowsize fftsize windowstep-oversamp 
                                     window-type normalize crossmode outfile)
  (let ((parampath (om::paramfile "tempcross.par"))
        (rep nil))
    (om::save-params param parampath)
    (setf rep (supervp-cross-synthesis sound1 sound2 parampath windowsize fftsize windowstep-oversamp 
                                       window-type normalize crossmode outfile))
    (delete-file parampath)
    rep))
